/*******************************
* ZK-40 web written by Katsumi *
*   This script is released    *
*     under the LGPL v2.1.     *
*******************************/
/*
CPU specifications:

Address: 12 bits
Data: 4 bits

Total 16384 bits (2048 bytes)

Resisters:
	A:	4 bits (Accumulator)
	B:	4 bits (General)
	C:	4 bits (General)
	H:	4 bits (Index high)
	L:	4 bits (Index low)
	F:	2 bits (flags: C and Z)
	PC:	11 bits (program counter)
	SP: 6 bits (stack pointer)

Segment resisters:
	S: 6 bits (upper of jump commands)

Mnemonic

0x00-0x07:	LD A,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x00     :	NOP
0x08-0x0f:	LD B,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x09     :	RET
0x10-0x17:	LD C,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x12     :	in A,(C)
0x18-0x1f:	LD (HL),x (x=A, B, C, (HL), H, L, Sh, Sl)
0x1b     :	out (C),A
0x20-0x27:	LD H,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x24     :	if Z
0x28-0x2f:	LD L,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x2d     :	if NZ
0x30-0x37:	LD Sh,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x36     :	if C
0x37     :  ???
0x38-0x3f:	LD Sl,x (x=A, B, C, (HL), H, L, Sh, Sl)
0x3e     :  ???
0x3f     :  if NC
0x40-0x47:	inc x (x=A, B, C, (HL), H, L, Sh, Sl)
0x48-0x4f:	dec x (x=A, B, C, (HL), H, L, Sh, Sl)
0x50-0x57:	push x (x=A, B, C, F, H, L, Sh, Sl)
0x58-0x5f:	pop x (x=A, B, C, F, H, L, Sh, Sl)
0x60-0x63:	ADD A,x (x=A, B, C, or (HL))
0x64-0x67:	SUB A,x (x=A, B, C, or (HL))
0x68-0x6b:	ADC A,x (x=A, B, C, or (HL))
0x6c-0x6f:	SBC A,x (x=A, B, C, or (HL))
0x70-0x73:	AND A,x (x=A, B, C, or (HL))
0x74-0x77:	OR  A,x (x=A, B, C, or (HL))
0x78-0x7b:	CP  A,x (x=A, B, C, or (HL))
0x7c-0x7f:	XOR A,x (x=A, B, C, or (HL))
0x80-0xbf:	LD S with 6 bits operand (skip flag does not change after this operation)
0xc0-0xff:	Jump/Call with 5 bits operand (MLB: Jump or Call)

Notes:

1) Skip flag may be set after "if" command.
If it is set, the next operation is skipped.

2) "LD S,xx" followed by "JP"/"CALL"/"LD x,Sl" can be recognized as 4 word commands.
To do this with "if" command, "LD S,xx" command does not change the status of skip flag.

3) "SL" (shift left) is the same as "ADD A".
"RL" (rotate left) is the same as "ADC A".
"SR" (shift right) is the same as "ADD A", "ADC A", "ADC A", and "ADC A".
"RR" (rotate right) is the same as ""ADC A", "ADC A", "ADC A", and "ADC A".

4) When loading to Sl, upper 2 bits will be set to 0.
On the other hand, when loading to Sh, lower 2 bits will be kept.

*/

/*
	Public methods:
	zk40cpu.reset();
	zk40cpu.setSpeed(clk);
	zk40cpu.exec(msec);
	
	Required outside methods:
	Methods written in z80functions.js.
	memory.read(addr);
	memory.write(addr,data);
	io.read(addr);
	io.write(addr,data);
	zk40cpu.events(); // optional
*/
zk40cpu=new Object();
zk40cpu.reset=function(){
	this.regA=0;
	this.regB=0;
	this.regC=0;
	this.regH=0;
	this.regL=0;
	this.regS=0;
	this.regF=0;
	this.regPC=0;
	this.regSP=0;
	this.skip=0;
};
zk40cpu.events=function(){
	// This will be called every msec.
};
zk40cpu.read=function(reg,flag){
	switch (reg&0x0f) {
		case 0x00:
			return this.regA;
		case 0x01:
			return this.regB;
		case 0x02:
			return this.regC;
		case 0x03:
			if (flag) {
				return this.regF;
			} else {
				//return memory.read(((this.regS & 0x3c)<<10) | (this.regH<<4) | this.regL);
				return memory.read(((this.regS & 0x0f)<<8) | (this.regH<<4) | this.regL);
			}
		case 0x04:
			return this.regH;
		case 0x05:
			return this.regL;
		case 0x06:
			return this.regS>>2;
		case 0x07:
		default:
			return this.regS&0x0f;
	}
};
zk40cpu.write=function(reg,val,flag){
	val&=0x0f;
	switch (reg&0x0f) {
		case 0x00:
			this.regA=val;
			return;
		case 0x01:
			this.regB=val;
			return;
		case 0x02:
			this.regC=val;
			return;
		case 0x03:
			if (flag) {
				this.regF=val;
				return;
			} else {
				//memory.write(((this.regS & 0x3c)<<10) | (this.regH<<4) | this.regL, val);
				memory.write(((this.regS & 0x0f)<<8) | (this.regH<<4) | this.regL, val);
				return;
			}
		case 0x04:
			this.regH=val;
			return;
		case 0x05:
			this.regL=val;
			return;
		case 0x06:
			this.regS=(val<<2)|(this.regS&0x03);
			return;
		case 0x07:
		default:
			this.regS=val;
			return;
	}
};
zk40cpu.push=function(data){
	this.regSP--;
	if (this.regSP<0) this.regSP=0x3f;
	memory.write(0xfc0|this.regSP,data&0x0f);
};
zk40cpu.pop=function(){
	var ret=memory.read(0xfc0|this.regSP);
	this.regSP++;
	if (0x3f<this.regSP) this.regSP=0;
	return ret&0x0f;
};
zk40cpu.special=function(code){
	switch (code&0x07) {
		case 0x00: // NOP
			return;
		case 0x01: // RET
			this.regPC=this.pop();
			this.regPC|=this.pop()<<4;
			this.regPC|=this.pop()<<8;
			return;
		case 0x02: // IN A,(C)
			this.regA=io.read(this.regC);
			return;
		case 0x03: // OUT (C),A
			io.write(this.regC,this.regA);
			return;
		case 0x04: // IF C
			this.skip=this.regF&0x01 ? false:true;
			return;
		case 0x05: // IF NC
			this.skip=this.regF&0x01 ? true:false;
			return;
		case 0x06: // IF Z
			this.skip=this.regF&0x02 ? false:true;
			return;
		case 0x07: // IF NZ
			this.skip=this.regF&0x02 ? true:false;
			return;
		default:
			
	}
}
zk40cpu.load=function(code){
	var dest=(code>>3)&0x07;
	var org=code&0x07;
	if (dest==org) {
		this.special(code);
	} else {
		this.write(dest,this.read(org));
	}
};
zk40cpu.setFlags=function(result){
	var flags;
	if (result&0x10) {
		// Set carry
		flags=0x01;
	} else {
		// Clear carry
		flags=0x00;
	}
	result&=0x0f;
	if (result==0x00) {
		// Set zero
		flags|=0x02;
	}
	this.write(3,flags,true);
};
zk40cpu.incDec=function(code){
	var result;
	if (code&0x08) {
		result=this.read(code&0x07)+0x1f;
	} else {
		result=this.read(code&0x07)+0x01;
	}
	this.setFlags(result);
	this.write(code&0x07,result&0x0f);
};
zk40cpu.pushPop=function(code){
	if (code&0x08) {
		this.write(code&0x07,this.pop(),true);
	} else {
		this.push(this.read(code&0x07,true));
	}
};
zk40cpu.calc=function(code){
	var result;
	switch (code&0x1c) {
		case 0x00: // ADD
			result=this.regA+this.read(code&0x03);
			break;
		case 0x04: // ADC
			result=this.regA+this.read(code&0x03);
			if (this.regF&0x01) result++;
			break;
		case 0x08: // SUB
		case 0x18: // CP
			result=this.regA-this.read(code&0x03);
			break;
		case 0x0c: // SBC
			result=this.regA-this.read(code&0x03);
			if (this.regF&0x01) result--;
			break;
		case 0x10: // AND
			result=this.regA&this.read(code&0x03);
			break;
		case 0x14: // OR
			result=this.regA|this.read(code&0x03);
			break;
		case 0x1c: // XOR
		default:
			result=this.regA^this.read(code&0x03);
			break;
	}
	result=(result+0x20)&0x1f;
	this.setFlags(result);
	if ((code&0x1c)!=0x18) this.write(0,result&0x0f);
};
zk40cpu.segment=function(code){
	this.regS=code&0x3f;
};
zk40cpu.jump=function(code){
	if (code&0x01) {
		// Call
		this.push(this.regPC>>8);
		this.push((this.regPC>>4)&0x0f);
		this.push(this.regPC&0x0f);
	}
	this.regPC=(this.regS<<6) | (code&0x3e);
};
zk40cpu.execOne=function(code){
	if (this.skip) {
		if (code<0x80 || 0xbf<code) this.skip=false;
		return;
	}
	switch (code&0xf0) {
		case 0x00: case 0x10: case 0x20: case 0x30:
			this.load(code);
			return;
		case 0x40:
			this.incDec(code);
			return;
		case 0x50:
			this.pushPop(code);
			return;
		case 0x60: case 0x70:
			this.calc(code);
			return;
		case 0x80: case 0x90: case 0xa0: case 0xb0:
			this.segment(code);
			return;
		default:
			this.jump(code);
			return;
	}
};
zk40cpu.speed=1000; // Default: 1 MHz
zk40cpu.setSpeed=function(clk){
	this.speed=Math.ceil(clk/1000);
};
zk40cpu.counter=1;
zk40cpu.exec=function(msec){
	var i,code;
	msec=msec*this.speed;
	for (i=0;i<msec;i++) {
		if (this.regPC==this.breakPoint) {
			this.breakPoint=-1;
			this.step=1;
			return;
		}
		code=memory.read(this.regPC++)<<4;
		code|=memory.read(this.regPC++);
		if (0xfff<this.regPC) this.regPC=0;
		this.execOne(code);
		if ((--this.counter)<1) {
			this.counter=this.speed;
			this.events();
		}
		if (this.step) return;
	}
};
